#ifndef _UTILS_CMD_
#define _UTILS_CMD_

/*
 * Header file for Monitoring Plugins utils_cmd.c
 *
 */
#include "../config.h"
#include <stddef.h>

/** types **/
typedef struct {
	char *buf;     /* output buffer */
	size_t buflen; /* output buffer content length */
	char **line;   /* array of lines (points to buf) */
	size_t lines;  /* lines of output */
} output;

/** prototypes **/
int cmd_run(const char *, output *, output *, int);
int cmd_run_array(char *const *, output *, output *, int);
int cmd_file_read(const char *, output *, int);

typedef struct {
	int error_code;
	int cmd_error_code;
	output out;
	output err;
} cmd_run_result;
cmd_run_result cmd_run2(const char *cmd, int flags);
cmd_run_result cmd_run_array2(char *const *cmd, int flags);

/* only multi-threaded plugins need to bother with this */
void cmd_init(void);
#define CMD_INIT cmd_init()

/* possible flags for cmd_run()'s fourth argument */
#define CMD_NO_ARRAYS 0x01 /* don't populate arrays at all */
#define CMD_NO_ASSOC  0x02 /* output.line won't point to buf */

void timeout_alarm_handler(int);

#endif /* _UTILS_CMD_ */
