#pragma once

#include "../../config.h"
#include "output.h"
#include "thresholds.h"
#include <mysql.h>

typedef struct {
	char *db_host;
	char *db_socket;
	char *db;
	char *db_user;
	char *db_pass;
	char *opt_file;
	char *opt_group;
	unsigned int db_port;

	char *sql_query;
	char *queryname;
	mp_thresholds thresholds;

	bool output_format_is_set;
	mp_output_format output_format;
} check_mysql_query_config;

check_mysql_query_config check_mysql_query_config_init() {
	check_mysql_query_config tmp = {
		.db_host = NULL,
		.db_socket = NULL,
		.db = NULL,
		.db_user = NULL,

		.db_pass = NULL,
		.opt_file = NULL,
		.opt_group = NULL,
		.db_port = MYSQL_PORT,

		.sql_query = NULL,
		.queryname = NULL,
		.thresholds = mp_thresholds_init(),

		.output_format_is_set = false,
	};
	return tmp;
}
